uniform vec3 BRICK_COLOR = vec3(0.69, 0.36, 0.20);
uniform vec3 MORTAR_COLOR = vec3(0.86, 0.75, 0.69);
uniform vec2 BRICK_SIZE = vec2(0.86, 0.26);
uniform vec2 BRICK_GAP = vec2(.1, .1);

varying vec3 N;
varying vec3 v;
varying vec2 vertexPosition;


void main()
{
	vec4 ambient, diffuse, specular;
	vec3 Normal, Viewer, color;
	vec2 fragPos;
	float diffuseShade, specularShade;

	// The scene's ambient light.
	ambient = gl_LightModel.ambient * gl_FrontMaterial.ambient;

	// The normal vectors is generally not normalized after being
	// interpolated across a triangle.  Here we normalize it.
	Normal = normalize(N);

	// Since the vertex is in eye space, the direction to the
	// viewer is simply the normalized vector from v to the
	// origin.
	Viewer = -normalize(v);

	// Compute factor to prevent light leakage from below the
	// surface
	float B = 1.0;
	if (dot(Normal, Viewer) < 0.0) B = 0.0;

	// Simulate a simple light emitting from the camera
	diffuseShade = max(dot(Normal, Viewer), 0.0);
	specularShade =
		B * pow(max(dot(normalize(Viewer + Viewer), Normal), 0.0), gl_FrontMaterial.shininess);

	fragPos = vertexPosition / BRICK_COLOR;

	// Offset alternating bricks
	if (frac(fragPos.y * 0.5) > 0.5) {
		fragPos.x += 0.5;
	}
	fragPos = frac(fragPos);

	// Select color based on fragment position relative to brick model
	float fractionalBrickWidth = BRICK_SIZE.x / (BRICK_SIZE.x + BRICK_GAP.x);
	float fractionalBrickHeight = BRICK_SIZE.y / (BRICK_SIZE.y + BRICK_GAP.y);

	if (fragPos.y <= fractionalBrickHeight
		&& fragPos.x <= fractionalBrickWidth) {
		color = BRICK_COLOR;	  // Fragment is within a brick
	}
	else {
		color = MORTAR_COLOR;  // Fragment is within the mortar
	}

	color *= diffuseShade + specularShade;

	// assign final color
	gl_FragColor = ambient + vec4(color, 1) + gl_FrontMaterial.emission;
}